// rm -fr viewer ; cc -Wall -lGLEW -lglut -lGLU -lX11 -L /usr/X11R6/lib -lGL -lm dyn_simpleGL_client.c -o viewer ; ./viewer 127.0.0.1 9021
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <GL/glew.h>
#include <GL/glut.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>


#define WINDOW_WIDTH 1000
#define WINDOW_HEIGHT 1000
#define MESH_WIDTH 1024
#define MESH_HEIGHT 1024
#define DS MESH_WIDTH * MESH_HEIGHT * 4 * sizeof(float)
#define FR 4

GLuint vbo;
float mesh[MESH_WIDTH * MESH_HEIGHT * 4];
int lastx,lasty;
int clicked=0;
int sockfd;

// mouse controls
int mouse_old_x, mouse_old_y;
int mouse_buttons = 0;
float rotate_x = 0.0, rotate_y = 0.0;
float translate_z = -3.0;

void keyboard( unsigned char key, int x, int y);
void mouse(int button, int state, int x, int y);
void motion(int x, int y);

/**********Network Functions************/
void connectToServer(int port, char* IP) {
   fprintf(stderr,"Attempting to connect to server...\n");
   struct sockaddr_in remote_addr;

   sockfd=socket(PF_INET,SOCK_STREAM,0);
   
   remote_addr.sin_family=AF_INET;
   remote_addr.sin_port=htons(port);
   remote_addr.sin_addr.s_addr=inet_addr(IP);
   memset(remote_addr.sin_zero, '\0',sizeof remote_addr.sin_zero);

   if(connect(sockfd,(struct sockaddr*)&remote_addr,sizeof remote_addr)==-1){
      perror("connect");
      exit(1);
   }
   
   fprintf(stderr,"Connected to server.\n");
}

void getNewFrame(void) {
//   fprintf(stderr,"Requesting new frame...\n");

   char newFrameRequest[17];
   newFrameRequest[0]=0x1;
   memset(&newFrameRequest[1], '\0',16);
   send(sockfd,newFrameRequest,17,0);

   int bytesRecieved=recv(sockfd,mesh,2*DS,MSG_PEEK);
   if(bytesRecieved==-1) { perror("recv()"); exit(1); }
//   fprintf(stderr,"Recieved frame. (%d bytes).\n",bytesRecieved);
}

void sendRequestToTerminate(void) {
   char terminate[17];
   terminate[0]=0x3;
   memset(&terminate[1],'\0',16);
   send(sockfd,terminate,17,0);
}
/***************************************/


/*************Display Functions*********/
void display(void) {
   int i;
   getNewFrame();  
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glTranslatef(0.0, 0.0, translate_z);
   glRotatef(rotate_x, 1.0, 0.0, 0.0);
   glRotatef(rotate_y, 0.0, 1.0, 0.0);

   glColor3f(1.0, 0.0, 0.0);
   glBegin(GL_POINTS);
    for(i = 0; i < MESH_WIDTH * MESH_HEIGHT; i+= 4)
    {
        glVertex3f(mesh[4*i], mesh[4*i+1], mesh[4*i+2]);
    }
   glEnd( );
   glDisableClientState(GL_VERTEX_ARRAY);

   glutSwapBuffers();
}

void idle(void) {
    glutPostRedisplay();
}

void keyboard( unsigned char key, int x, int y) {
    switch( key) {
        case 27:
        sendRequestToTerminate();
        exit (0);
        default: break;
    }
}


/**************************************************/

void sendEnvironmentVariables(void) {
   fprintf(stderr,"Sending environment data...\n");
   int dim=MESH_WIDTH;
   send(sockfd,&dim,sizeof(int),0);
   fprintf(stderr,"Sent environment data.\n");

}

void cleanup(void) {
   close(sockfd);
}

void initGL()
{
    // initialize necessary OpenGL extensions
    glewInit();

    // default initialization
    glClearColor( 0.0, 0.0, 0.0, 1.0);
    glDisable( GL_DEPTH_TEST);

    // viewport
    glViewport( 0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);

    // projection
    glMatrixMode( GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(60.0, (GLfloat)WINDOW_WIDTH / (GLfloat) WINDOW_HEIGHT, 0.1, 10.0);
}

int main(int argc, char** argv) {

   if(argc!=3) {
      fprintf(stderr,"Usage: %s <IP address> <port>\n",argv[0]);
      exit(1);
   }

   connectToServer(atoi(argv[2]),argv[1]);
   sendEnvironmentVariables();

   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
   glutInitWindowSize(WINDOW_WIDTH, WINDOW_HEIGHT);
   glutCreateWindow("Compute Stable Fluids");
   glutDisplayFunc(display);
   glutKeyboardFunc(keyboard);
   glutMouseFunc(mouse);
   glutMotionFunc(motion);
   glutIdleFunc(idle);

   initGL( );
   atexit(cleanup);

   glutMainLoop();

   return 0;
}
//
////////////////////////////////////////////////////////////////////////////////
//! Mouse event handlers
////////////////////////////////////////////////////////////////////////////////
void mouse(int button, int state, int x, int y)
{
    if (state == GLUT_DOWN) {
        mouse_buttons |= 1<<button;
    } else if (state == GLUT_UP) {
        mouse_buttons = 0;
    }

    mouse_old_x = x;
    mouse_old_y = y;
    glutPostRedisplay();
}

void motion(int x, int y)
{
    float dx, dy;
    dx = x - mouse_old_x;
    dy = y - mouse_old_y;

    if (mouse_buttons & 1) {
        rotate_x += dy * 0.2;
        rotate_y += dx * 0.2;
    } else if (mouse_buttons & 4) {
        translate_z += dy * 0.01;
    }

    mouse_old_x = x;
    mouse_old_y = y;
}
